/*
    FEARLUS model0-6-8-2: ClumpinessReport.m
    Copyright (C) 1999-2004, 2007-2008  Macaulay Institute

    This file is part of FEARLUS model0-6-8-2, an agent-based model of land use
    change.

    FEARLUS model0-6-8-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-8-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


The implementation for the ClumpinessReport Object

*/

#import "ClumpinessReport.h"
#import "Parameter.h"
#import "ModelSwarm.h"
#import "FearlusArguments.h"
#import "Environment.h"
#import "BitString.h"
#import "MiscFunc.h"
#import <collections/Array.h>
#import <defobj/Zone.h>
#import "LandParcel.h"
#import "Number.h"
#import "HashTableFEARLUS.h"
#import <stdlib.h>
#import <errno.h>

@implementation ClumpinessReport

/*

create:

Creation method. 

*/

+(id <Report>)create: aZone {
  ClumpinessReport *r;

  r = [super create: aZone];
  r->histoflag = 0;

  return r;
}

/*

setModelSwarm:andParameters:

Set the model swarm and parameters for this report. These will be used to
obtain the data in the report. 

*/

-(void)setModelSwarm: (ModelSwarm *)m andParameters: (Parameter *)p {
  parameter = p;
  model = m;
}

/*

reportForYear:toFile:

Create a report showing the median number of bits in the land use bitstring
that match between Moore neighbours in the environment.

Note that [FearlusArguments nl] is used as the string to print for a
new line. This is in case DOS mode has been specified on the command
line.

*/

-(void)reportForYear: (unsigned)year toFile: (FILE *)fp {
  id <Array> histo;
  double median, mean;
  int b, x, y, xx, yy, xxx, yyy, xsize, ysize, lpbssize;
  int nlinks = 0;
  char *buf;
  int bufsz, bufsz10;
  double total;
  double halftotal;
  Environment *env;
  id <Zone> histoZone = [Zone create: scratchZone];

  env = [model getEnvironment];
  xsize = (int)[parameter environmentXSize];
  ysize = (int)[parameter environmentYSize];
  lpbssize = (int)[parameter parcelBitStringSize];

  histo = [Array create: histoZone setCount: lpbssize + 1];

  for(b = 0; b <= lpbssize; b++) {
    Number *n = [Number create: histoZone];

    [n setInt: 0];
    [histo atOffset: b put: n];
  }
    

  /* Loop through the land parcels and their Moore neighbours and build up
     the histogram of the distribution of matching bitstrings.
   */

  if(![HashTableFEARLUS createHashEstimatedSize: xsize * ysize * 4]) {
    fprintf(stderr, "ClumpinessReport: Could not create hash table\n");
    exit(1);
  }
  bufsz = 6;
  for(bufsz10 = 1; xsize > bufsz10 && ysize > bufsz10; bufsz10 *= 10) {
    bufsz += 4;
  }
  buf = (char *)malloc(bufsz * sizeof(char));
  if(buf == NULL) {
    perror("ClumpinessReport: Memory allocation");
    exit(1);
  }
  for(x = 0; x < xsize; x++) {
    for(y = 0; y < ysize; y++) {
      BitString *bs1 = [[env getLandParcelAtX: x Y: y] getAttributes];
      for(xx = -1; xx <= 1; xx++) {
	for(yy = -1; yy <= 1; yy++) {
	  BitString *bs2;
	  int match;
	  int retval;
	  int i;
	  Number *n, *m;

	  if(xx == 0 && yy == 0) continue;
	  xxx = [MiscFunc Mod: (x + xx) inRange0To: xsize];
	  yyy = [MiscFunc Mod: (y + yy) inRange0To: ysize];
	  if(x == xxx && y == yyy) continue;
				// Don't want to test the clumpiness of a cell
				// with itself...
	  bs2 = [[env getLandParcelAtX: xxx Y: yyy] getAttributes];
	  match = (int)[bs1 getMatchWith: bs2];
	  m = [Number create: histoZone];
	  [m setInt: match];
	  for(i = 0; i < bufsz; i++) {
	    buf[i] = '\0';
	  }
	  if((x < xxx) || (x == xxx && y < yyy)) {
	    sprintf(buf, "%d,%d:%d,%d", x, y, xxx, yyy);
	  }
	  else {
	    sprintf(buf, "%d,%d:%d,%d", xxx, yyy, x, y);
	  }
	  retval = [HashTableFEARLUS addKey: buf data: m];
	  if(retval < 0) {
	    fprintf(stderr, "ClumpinessReport: Ooops, hash table full -- "
		    "should have configured it with more elements...\n");
	  }
	  else if(retval == 0) {
	    [m drop];
	  }
	  else {
	    n = (Number *)[histo atOffset: match];
	    [n setInt: [n getInt] + 1];
	    nlinks++;
	  }
	}
      }
    }
  }
  free(buf);
  [HashTableFEARLUS destroyHash];
 
  /* Find the median and the mean */

  halftotal = (double)nlinks / 2.0;
  total = 0.0;
  mean = 0.0;
  median = -1.0;		// Use a negative median to indicate it hasn't
				// been set. This is OK because the median
				// wouldn't ever be negative in this
				// distribution.
  for(b = 0; b <= lpbssize; b++) {
    double nmatch = (double)[(Number *)[histo atOffset: b] getInt];

    total += nmatch;
    mean += (double)b * (nmatch / (double)nlinks);
    if(median < 0.0) {		// Negative median means it hasn't been set yet
      if(total == halftotal) {
	median = (double)b + 0.5;
      }
      else if(total > halftotal) {
	median = (double)b;
      }
    }
  }

  /* Print the results */

  if(histoflag) {
    fprintf(fp, "Number of matching bits\tFrequency\tFrequency as proportion"
	    "%s", [FearlusArguments nl]);
    for(b = 0; b <= lpbssize; b++) {
      fprintf(fp, "%d\t%d\t%.16g%s", b, [(Number *)[histo atOffset: b] getInt],
	      (double)[(Number *)[histo atOffset: b] getInt] / (double)nlinks,
	      [FearlusArguments nl]);
    }
  }
  fprintf(fp, "Median number of matching bits\t%g%s", median,
	  [FearlusArguments nl]);
  fprintf(fp, "Median proportion of matching bits\t%g%s",
	  median / (double)lpbssize, [FearlusArguments nl]);
  fprintf(fp, "Mean number of matching bits\t%g%s", mean,
	  [FearlusArguments nl]);
  fprintf(fp, "Mean proportion of matching bits\t%g%s",
	  mean / (double)lpbssize, [FearlusArguments nl]);

 
  /* Clear up */

  [histoZone drop];
}
/*

setOption:toValue:

setOptionFlag:toValue:

The histo option can be given as Histo, or NoHisto to unset (default).

*/

-(BOOL)setOption: (char *)option toValue: (char *)value {
  fprintf(stderr, "Invalid format in report configuration file %s\n"
	  "%s: There are no options for this report: (%s=%s).\n",	
	  [FearlusArguments getReporterCFFile], class_get_class_name([self class]),
	  option, value);
  exit(1);
}

-(BOOL)setOptionFlag: (char *)option toValue: (BOOL)value {
  if(strcmp(option, "Histo") == 0) {
    if(value) {
      histoflag = YES;
    }
    else {
      histoflag = NO;
    }
  }
  else {
    if(value) {
      fprintf(stderr, "Invalid format in report configuration file %s\n"
	      "%s: Option flag not recognised: (%s).\n",	
	      [FearlusArguments getReporterCFFile],
	      class_get_class_name([self class]),
	      option);
      exit(1);
    }
    else {
      fprintf(stderr, "Invalid format in report configuration file %s\n"
	      "%s: Option flag not recognised: (No%s).\n",	
	      [FearlusArguments getReporterCFFile],
	      class_get_class_name([self class]),
	      option);
      exit(1);
    }
  }
  return YES;
}

/*

setFixedYears:

Set the fixed years for reporting.

*/

-(void)setFixedYears: (id <List>)y {
  fixedYears = y;
}

/*

setRepeatYears:

Set the repeating years for reporting.

*/

-(void)setRepeatYears: (id <List>)y {
  repeatYears = y;
}

/* There's no need for a drop method -- the Reporter object will own all Report
   objects, and stores them in its own Zone. When the zone is deleted, all the
   Reports' allocated memory will go with it, so long as we use [self getZone]
   here when allocating memory. */

/*

reportingForYear:

Return YES if we're reporting this year, according to the years we've been
handed. NO otherwise.

*/

-(BOOL)reportingForYear: (unsigned)year {
  BOOL reportingYear;
  id inx;
  Number *n;

  /* Decide if we should print this year */
  reportingYear = NO;
  if(([fixedYears getCount] > 0)
     && ([(Number *)[fixedYears atOffset: 0] getUnsigned] == year)) {
    /* This is a fixed year to print. So we'll go ahead, but after we've
       removed the head */
    [[fixedYears removeFirst] drop];
    reportingYear = YES;
  }
  else {
    /* Not a fixed year to print, so check all the repeating years */
    for(inx = [repeatYears begin: scratchZone], [inx next];
	[inx getLoc] == Member;
	[inx next]) {
      n = [inx get];
      if((year % [n getUnsigned]) == 0) {
	/* It's a repeating year to print */
	reportingYear = YES;
	break;
      }
    }
    [inx drop];
  }
  return reportingYear;
}

@end

