/*
    FEARLUS model0-6-8-2: LandAllocator.m
    Copyright (C) 1999-2004, 2007-2008  Macaulay Institute

    This file is part of FEARLUS model0-6-8-2, an agent-based model of land use
    change.

    FEARLUS model0-6-8-2 is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    FEARLUS model0-6-8-2 is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details. (LICENCE file in
    this directory.)

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    Contact information:
      Gary Polhill,
      Macaulay Institute, Craigiebuckler, Aberdeen, AB15 8QH. United Kingdom
      g.polhill@macaulay.ac.uk
*/
/*


Implementation of the LandAllocator class.

*/

#import "LandAllocator.h"
#import "LandManager.h"
#import "Environment.h"
#import "LandParcel.h"
#import "FearlusArguments.h"
#import "Parameter.h"
#import "SubPopulation.h"
#import "MiscFunc.h"		// shuffleList:

@implementation LandAllocator


/*

setMinColour:Max:

Set the maximum and minimum colours that can be allocated to land managers.
This is an observer thing, I know, and hence bad practice, but we'll have to
live with it, because I can't think of a better way to do it.

*/

-(void)setMinColour: (int)min max: (int)max {
  minColour = min;
  maxColour = max;
}

/*

setParametersInitStrategyMax:initStrategyMin:strategyUnit:
landUseChangeYieldThreshold:transferLandUseYieldThreshold:neighbourMutateMax:
neighbourMutateMin:

Set the model parameters appropriate to the land allocator.

*/

-(void)setParameters: (Parameter *)p {
  parameter = p;
  landManagers = [List create: [self getZone]];
  deadLandManagers = [List create: [self getZone]];
  justDeadLandManagers = [List create: [self getZone]];
  deadDeadLandManagers = [List create: [self getZone]];
  parcelsForSale = [List create: [self getZone]];
}

/*

initialiseWithEnvironment

Method to initialise the land allocator object. Passes in the
environment. Sets up the land manager lists, creates the land
managers.

*/

-(void)initialiseWithEnvironment: (Environment *)e {
  int maxx, maxy, x, y;
  unsigned nlpx, nlpy;
  LandParcel *lp;
  LandManager *lm;

  // Set all the parameters for the land allocator
  nextColour = minColour;
  environment = e;
  // For each cell in the environment, create a land manager to manage it
  maxx = [environment getSizeX];
  maxy = [environment getSizeY];
  nlpx = [parameter nInitXParcels];
  nlpy = [parameter nInitYParcels];
  if(maxx % nlpx != 0) {
    fprintf(stderr, "Environment X dimension %d not divisible by the "
	    "number of initial land parcels on the X axis to allocate to land "
	    "managers %u\n", maxx, nlpx);
    exit(1);
  }
  if(maxy % nlpy != 0) {
    fprintf(stderr, "Environment Y dimension %d not divisible by the "
	    "number of initial land parcels on the Y axis to allocate to land "
	    "managers %u\n", maxy, nlpy);
    exit(1);
  }
  for(x = 0; x < maxx / nlpx; x++) {
    for(y = 0; y < maxy / nlpy; y++) {
      int x1, y1;

      lm = [self createNewLandManager];
      if([FearlusArguments debugAllocator]) {
	printf("A: Created land manager ID %u\n", [lm getPIN]);
      }
      for(x1 = 0; x1 < nlpx; x1++) {
	for(y1 = 0; y1 < nlpy; y1++) {
	  lp = [environment getObjectAtX: ((x * nlpx) + x1)
			    Y: ((y * nlpy) + y1)];
	  [lm addLandParcel: lp];	// Allocate the parcel to the manager
	  [lp setLandManager: lm];	// Tell the parcel who its manager is
	  if([FearlusArguments debugAllocator]) {
	    printf("A: Assigned land parcel %u at (%d, %d) to land manager"
		   "ID %u\n", [lp getPIN], ((x * nlpx) + x1),
		   ((y * nlpy) + y1), [lm getPIN]);
	  }
	}
      }
    }
  }
}

/*

addLandParcelForSale:

Add a land parcel to the list of land parcels for sale. This method is called
from land managers when they are selling off land parcels in order to get out
of the red.

*/

-(void)addLandParcelForSale: (LandParcel *)lp {
  [parcelsForSale addLast: lp];
}

/*

transferLandParcels

First shuffle the list of land parcels for sale using a list-shuffler
object.  Then find land managers to buy each of the land parcels. A
Land Manager may buy a land parcel if they have enough wealth to cover
the landParcelPrice without going into the red. There is always a
chance of creating a new land manager to own the parcel.

*/

-(void)transferLandParcels {
  id elml, tempZone;
  LandManager *lm;
  LandParcel *lp, *nlp;
  int nelm, selectlm;

  tempZone = [Zone create: scratchZone];

  if([parcelsForSale getCount] == 0) {
    if([FearlusArguments debugAllocator]) {
      printf("A: No land parcels to transfer\n");
    }
    return;
  }
  // Shuffle the list of land parcels for sale
  [MiscFunc shuffleList: parcelsForSale];
  // I'd love to use the Swarm built-in ListShuffler to do this, but it
  // didn't work.

  elml = [List create: tempZone];
				// A list to store land managers who are
				// eligible to receive new land parcels
  // Loop through the land parcels to allocate them to new land managers

  while([parcelsForSale getCount] > 0) {
    id <Index> nlpix;

    lp = [parcelsForSale removeFirst];

    for(nlpix = [lp nbrBegin: scratchZone], nlp = (LandParcel *)[nlpix next];
	[nlpix getLoc] == Member;
	nlp = (LandParcel *)[nlpix next]) {
				// Loop through the neighbours of the parcel
      lm = [nlp getLandManager];
      if([lm eligibleForNewLandParcels]) {
	[elml addLast: lm];
	if([FearlusArguments debugCoreMethods]) {
	  printf("A: %s: Adding land manager %u, owner of parcel %u at (%d, "
		 "%d) to the list of managers eligible to buy parcel %u at "
		 "(%d, %d)\n", sel_get_name(_cmd), [lm getPIN], [nlp getPIN],
		 [nlp getX], [nlp getY], [lp getPIN], [lp getX], [lp getY]);
	}
      }
    }
    [nlpix drop];
    nelm = (int)[elml getCount];
    selectlm = [uniformIntRand getIntegerWithMin: 0 withMax: nelm];
				// Choose a land manager
    if([FearlusArguments debugCoreMethods]) {
      int i;

      printf("A: %s: Choice of manager to buy parcel %u at (%d, %d) made as "
	     "follows:\n", sel_get_name(_cmd), [lp getPIN], [lp getX],
	     [lp getY]);
      printf("\tManager\tTicket\n");
      for(i = 0; i < nelm; i++) {
	printf("\t%7u\t% 6d", [[elml atOffset: i] getPIN], i);
	if(selectlm == i) {
	  printf(" <-- Choice = %d", selectlm);
	}
	printf("\n");
      }
      printf("\t New LM\t% 6d", nelm);
      if(selectlm == nelm) {
	printf(" <-- Choice = %d", nelm);
      }
      printf("\n");
    }
    if(selectlm == nelm) {	// Chosen a new land manager
      lm = [self createNewLandManager];
      if([FearlusArguments debugAllocator]) {
	printf("A: Created new land manager ID %u for land parcel %u at "
	       "(%d, %d)\n", [lm getPIN], [lp getPIN], [lp getX], [lp getY]);
      }
				// Create the new land manager
    }
    else {			// Chosen an existing land manager
      lm = [elml atOffset: selectlm];
				// Get the land manager from the eligibility
				// list
    }
    [lm addLandParcel: lp];	// Allocate the land parcel to the land mgr
    if([FearlusArguments debugAllocator]) {
      printf("A: Transferred land parcel %u at (%d, %d) from land manager %u "
	     "to land manager %u\n", [lp getPIN], [lp getX], [lp getY],
	     [[lp getLandManager] getPIN], [lm getPIN]);
    }
    [lp setNextLandManagerTo: lm];
				// Tell the land parcel who its land mgr will
				// be at the start of the next year
    [elml removeAll];		// Clear the list of eligible land mgrs ready
				// for the next land parcel
  }
  [tempZone drop];
}

/*

createNewLandManager

Create a new land manager, initialise it, and store it in the list of land
managers

*/

-(LandManager *)createNewLandManager {
  LandManager *lm = nil;
  double choice;		// Choice of sub-population
  double cumulativeProb;
  unsigned i;

				// Create a new land manager
  // Decide the sub-population of the land manager
  cumulativeProb = 0.0;
  choice = [uniformDblRand getDoubleWithMin: 0.0 withMax: 1.0];
  for(i = 0; i < [parameter nSubPopulations]; i++) {
    SubPopulation *sp = [parameter subPopulation: i];
    cumulativeProb += [sp getProb];
    if(choice <= cumulativeProb) {
      lm = [[sp getLandManagerClass] create: [self getZone]];
      [lm setSubPopulation: sp];
      [sp addLandManager: lm];
      break;
    }
  }
  if(choice > cumulativeProb) {
    fprintf(stderr, "%s -- Probabilities for sub-populations do not sum to 1.0"
	    " (%g). Aborting simulation.\n", sel_get_name(_cmd), 
	    cumulativeProb);
    abort();
  }
  // If we're debugging core methods, then print a message showing how the
  // choice was made.
  if([FearlusArguments debugCoreMethods]) {
    printf("A: %s: Sub-population of new land manager %u selected as "
	   "follows:\n\tSub-pop Probability (Cumulative)\n",
	   sel_get_name(_cmd), [lm getPIN]);
    cumulativeProb = 0.0;
    for(i = 0; i < [parameter nSubPopulations]; i++) {
      cumulativeProb += [[parameter subPopulation: i] getProb];
      printf("\t%7u       %g (%g)", [[parameter subPopulation: i] getPIN],
	     [[parameter subPopulation: i] getProb], cumulativeProb);
      if([parameter subPopulation: i] == [lm getSubPopulation]) {
	printf(" <-- choice = %g\n", choice);
      }
      else {
	printf("\n");
      }
    }
  }
  [lm setParameters: parameter];
  [lm initialiseWithEnvironment: environment
      landAllocator: self
      colour: [self getNextFreeColour] ];
				// Initialise the land manager
  [landManagers addLast: lm];
  barnCredit += [parameter nInitXParcels] // barn should be wealth
    * [parameter nInitYParcels] * [parameter landParcelPrice];
  // Each land manager created gets their first land parcel free. Keep track of
  // the amount given to land managers. (This won't just be landParcelPrice *
  // numberOfLandManagersCreated because we'll also be subtracting any dead
  // land manager's final wealth from the barnCredit.)
  return(lm);
}

/*

getLandManagers

Return the list of land managers (who are currently in the game)

*/

-getLandManagers {
  return(landManagers);
}

/*

killLosers

Rather a brutal name for a method that moves land managers who have no land
parcels left onto the list of land managers who are no longer making decisions
about land parcels. Works by removing all land managers from the land manager
list one by one, and sorting them into two lists according to whether or not
they are staying in the game. Then copy the list of managers staying in the
game back into the land manager list.

*/

-(void)killLosers {
  id tempZone, lml;
  LandManager *lm;
  int nKilled;

  tempZone = [Zone create: scratchZone];
  lml = [List create: tempZone];
				// A list that will store land managers to
				// stay in the game
  // Loop through all the land managers sorting them into winners and losers
  nKilled = 0;
  while([landManagers getCount] > 0) {
    lm = [landManagers removeFirst];
    if([lm eligibleForRemovalFromGame]) {
      [deadLandManagers addLast: lm];
				// Remove the land manager from the game
      [justDeadLandManagers addLast: lm];
      // (^) remember the land managers who died this year so they can record
      //     their history. (Assume, therefore, that killLosers occurs before
      //     recordHistory in the schedule.)
      nKilled++;
      if([FearlusArguments debugAllocator]) {
	if(nKilled == 1) {
	  printf("A: Killing land manager(s): %u", [lm getPIN]);
	}
	else {
	  printf(", %u", [lm getPIN]);
	}
      }
      [lm kill];		// Tell the land manager they are out
      barnCredit -= [lm getBarn]; // should be wealth
      // Subtract the final wealth of the land manager from the barnCredit
      // -- effectively we are keeping track of the amount of wealth the land
      // allocator would have to pay to balance the books, as it were.
      [[lm getSubPopulation] removeLandManager: lm];
    }
    else {
      [lml addLast: lm];	// Keep this land manager
    }
  }
  if([FearlusArguments debugAllocator]) {
    if(nKilled == 0) {
      printf("A: No land managers to kill\n");
    }
    else {
      printf("\n");
    }
  }
  // Loop through the list of winners and add them back to the list of land
  // managers that are in the game.
  while([lml getCount] > 0) {
    lm = [lml removeFirst];
    [landManagers addLast: lm];
  }
  [lml drop];
  [tempZone drop];
}

/*

getNextFreeColour

For now, this method is a simple loop round the available colours from min to
max. In future, this should (a) ensure that sequential colours are as far apart
in colour space as possible (b) keep a record of dead land manager's colours, 
and assign them out.

*/

-(int)getNextFreeColour {
  if(++nextColour > maxColour) {
    nextColour = minColour;
  }
  return nextColour;
}

/*

writeHistory

Write the history of the land allocator and associated objects to a file

*/

-(void)writeHistory: (FILE *)fp {
  id iterator;
  LandManager *lm;

  fprintf(fp, "Type\tID\tSub-Pop ID\tYear\tParcels Gained\tAge\tBarn\t"
	  "Neighbourhood Weight\tNeighbour Information Loss\t"
	  "Habit Imitation Yield Threshold\tProbability of Imitative Strategy "
	  "Method\tInitial Strategy\tAbove Habit Threshold Strategy\t"
	  "Below Habit Threshold Imitative Strategy\t"
	  "Below Habit Threshold Non-Imitative Strategy%s", [FearlusArguments nl]);
  for(iterator = [landManagers begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lm = [iterator get];
    [lm printHistoryToFile: fp];
  }
  [iterator drop];
  /* Repeat for dead land managers */
  for(iterator = [deadLandManagers begin: scratchZone], [iterator next];
      [iterator getLoc] == Member;
      [iterator next]) {
    lm = [iterator get];
    [lm printHistoryToFile: fp];
    [deadDeadLandManagers addLast: lm];
  }
  [iterator drop];
  [deadLandManagers removeAll];
}

/*

recordHistory

Record the history (live land managers only!)

*/

-(void)recordHistory {
  [landManagers forEach: M(recordHistory)];
  [justDeadLandManagers forEach: M(recordHistory)];
  [justDeadLandManagers removeAll];
}

/*

getBarnCredit (should be wealth)

Return the barnCredit as a double (because I reckon the graphing tools probably
can't handle long longs). The barnCredit is the amount of wealth effectively
given to land managers through providing them with their initial land parcel
and paying off their debts when they die. It can be reduced if land managers
die in credit, which is possible for certain settings of the landParcelPrice.

*** The barnCredit is now a double anyway. The landParcelPrice and break even
threshold can both be non-integral (and still make sense). ***

*/

-(double)getBarnCredit {	// wealth
  return barnCredit;		// wealth
}

/*

getWealth

Return the sum of the wealths of the land managers minus the barnCredit, as a
double so that it can be graphed. This represents the amount of wealth created
during the simulation.

*/

-(double)getWealth {
  id lmi;
  double total;

  total = 0;
  for(lmi = [landManagers begin: scratchZone], [lmi next];
      [lmi getLoc] == Member;
      [lmi next]) {
    total += [[lmi get] getBarn]; // wealth
  }
  [lmi drop];
  total -= barnCredit;		// wealth
  return total;
}

/*

drop

Delete all items created by this object

*/

-(void)drop {
  [landManagers deleteAll];
  [deadLandManagers deleteAll];
  [landManagers drop];
  [deadLandManagers drop];
  [super drop];
}

@end
